'use strict';
Object.defineProperty(exports, "__esModule", { value: true });
exports.EndpointGenerator = void 0;
var lodash_1 = require("lodash");
var typescript_1 = require("typescript");
var debug_1 = require("../debug");
var type_1 = require("../decorator/type");
var utils_1 = require("../decorator/utils");
var pathUtils_1 = require("../utils/pathUtils");
var resolver_1 = require("./resolver");
var EndpointGenerator = /** @class */ (function () {
    // -------------------------------------------
    function EndpointGenerator(node, current) {
        this.current = current;
        this.debugger = debug_1.useDebugger();
        this.node = node;
    }
    // -------------------------------------------
    // --------------------------------------------------------------------
    EndpointGenerator.prototype.generatePath = function (key) {
        var values = [];
        var handler = type_1.Decorator.getRepresentationHandler(key, this.current.decoratorMap);
        var config = handler.buildRepresentationConfigFromNode(this.node);
        var property = handler.getPropertyByType(config.name);
        if (typeof property !== 'undefined') {
            var argument = handler.getDecoratorPropertyValueAsItem(config.decorator, property);
            if (typeof argument !== 'undefined') {
                values.push(argument);
            }
        }
        this.path = pathUtils_1.normalizePath(values.join('/'));
    };
    // --------------------------------------------------------------------
    EndpointGenerator.prototype.getDecoratorValues = function (decoratorName, acceptMultiple) {
        if (acceptMultiple === void 0) { acceptMultiple = false; }
        var decorators = utils_1.getDecorators(this.node, function (decorator) { return decorator.text === decoratorName; });
        if (!decorators || !decorators.length) {
            return [];
        }
        if (!acceptMultiple && decorators.length > 1) {
            throw new Error("Only one " + decoratorName + " decorator allowed in " + this.getCurrentLocation() + ".");
        }
        var result;
        if (acceptMultiple) {
            result = decorators.map(function (d) { return d.arguments; });
        }
        else {
            var d = decorators[0];
            result = d.arguments;
        }
        this.debugger('Arguments of decorator %s: %j', decoratorName, result);
        return result;
    };
    // -------------------------------------------
    EndpointGenerator.prototype.getSecurity = function () {
        var _this = this;
        var securities = this.getDecoratorValues('Security', true);
        if (!securities || !securities.length) {
            return undefined;
        }
        return securities.map(function (security) { return ({
            name: security[1] ? security[1] : 'default',
            scopes: security[0] ? lodash_1.castArray(_this.handleRolesArray(security[0])) : []
        }); });
    };
    EndpointGenerator.prototype.handleRolesArray = function (argument) {
        if (typescript_1.isArrayLiteralExpression(argument)) {
            return argument.elements.map(function (value) { return value.getText(); })
                .map(function (val) { return (val && val.startsWith('\'') && val.endsWith('\'')) ? val.slice(1, -1) : val; });
        }
        else {
            return argument;
        }
    };
    // -------------------------------------------
    EndpointGenerator.prototype.getExamplesValue = function (argument) {
        var _this = this;
        var example = {};
        if (typeof argument === 'undefined') {
            return example;
        }
        this.debugger(argument);
        if (argument.properties) {
            argument.properties.forEach(function (p) {
                example[p.name.text] = _this.getInitializerValue(p.initializer);
            });
        }
        else {
            example = this.getInitializerValue(argument);
        }
        this.debugger('Example extracted for %s: %j', this.getCurrentLocation(), example);
        return example;
    };
    EndpointGenerator.prototype.getInitializerValue = function (initializer) {
        var _this = this;
        switch (initializer.kind) {
            case typescript_1.SyntaxKind.ArrayLiteralExpression:
                return initializer.elements.map(function (e) { return _this.getInitializerValue(e); });
            case typescript_1.SyntaxKind.StringLiteral:
                return initializer.text;
            case typescript_1.SyntaxKind.TrueKeyword:
                return true;
            case typescript_1.SyntaxKind.FalseKeyword:
                return false;
            case typescript_1.SyntaxKind.NumberKeyword:
            case typescript_1.SyntaxKind.FirstLiteralToken:
                return parseInt(initializer.text, 10);
            case typescript_1.SyntaxKind.ObjectLiteralExpression:
                var nestedObject_1 = {};
                initializer.properties.forEach(function (p) {
                    nestedObject_1[p.name.text] = _this.getInitializerValue(p.initializer);
                });
                return nestedObject_1;
            default:
                return undefined;
        }
    };
    // -------------------------------------------
    EndpointGenerator.prototype.getResponses = function () {
        var _this = this;
        var handler = type_1.Decorator.getRepresentationHandler('RESPONSE_DESCRIPTION', this.current.decoratorMap);
        var config = handler.buildRepresentationConfigFromNode(this.node);
        var args = handler.getPropertiesByTypes(config.name, ['STATUS_CODE', 'DESCRIPTION', 'PAYLOAD', 'TYPE']);
        if (!config.decorators || !config.decorators.length) {
            return [];
        }
        this.debugger('Generating Responses for %s', this.getCurrentLocation());
        return config.decorators.map(function (decorator) {
            var description = handler.getDecoratorPropertyValueAsItem(decorator, args['DESCRIPTION']) || 'Ok';
            var status = handler.getDecoratorPropertyValueAsItem(decorator, args['STATUS_CODE']) || '200';
            var examples = handler.getDecoratorPropertyValueAsItem(decorator, args['PAYLOAD']);
            if (typeof examples !== 'undefined') {
                examples = _this.getExamplesValue(examples);
            }
            var type = handler.getDecoratorPropertyValueAsItem(decorator, args['TYPE']);
            var responses = {
                description: description,
                examples: examples,
                schema: type ? new resolver_1.TypeNodeResolver(type, _this.current).resolve() : undefined,
                status: status
            };
            _this.debugger('Generated Responses for %s: %j', _this.getCurrentLocation(), responses);
            return responses;
        });
    };
    // -------------------------------------------
    EndpointGenerator.prototype.getProduces = function () {
        var handler = type_1.Decorator.getRepresentationHandler('RESPONSE_PRODUCES', this.current.decoratorMap);
        var config = handler.buildRepresentationConfigFromNode(this.node);
        var produces = handler.getDecoratorPropertyValueAsArray(config.decorator, handler.getPropertyByType(config.name));
        if (typeof produces === 'undefined' || produces.length === 0) {
            var acceptHandler = type_1.Decorator.getRepresentationHandler('REQUEST_ACCEPT', this.current.decoratorMap);
            var acceptConfig = acceptHandler.buildRepresentationConfigFromNode(this.node);
            return acceptHandler.getDecoratorPropertyValueAsArray(acceptConfig.decorator, acceptHandler.getPropertyByType(acceptConfig.name));
        }
        return produces;
    };
    EndpointGenerator.prototype.getConsumes = function () {
        var handler = type_1.Decorator.getRepresentationHandler('REQUEST_CONSUMES', this.current.decoratorMap);
        var config = handler.buildRepresentationConfigFromNode(this.node);
        return handler.getDecoratorPropertyValueAsArray(config.decorator, handler.getPropertyByType(config.name));
    };
    EndpointGenerator.prototype.getTags = function () {
        var handler = type_1.Decorator.getRepresentationHandler('SWAGGER_TAGS', this.current.decoratorMap);
        var config = handler.buildRepresentationConfigFromNode(this.node);
        return handler.getDecoratorPropertyValueAsArray(config.decorator, handler.getPropertyByType(config.name));
    };
    return EndpointGenerator;
}());
exports.EndpointGenerator = EndpointGenerator;
//# sourceMappingURL=endpointGenerator.js.map